package qrzws

import (
	"fmt"
	"net/http"
	"regexp"
	"strings"

	"git.paulbsd.com/paulbsd/qrz/src/config"
	"git.paulbsd.com/paulbsd/qrz/src/page"
	"git.paulbsd.com/paulbsd/qrz/src/qrz"
	"git.paulbsd.com/paulbsd/qrz/src/static"
	"git.paulbsd.com/paulbsd/qrz/src/templates"
	"github.com/gobuffalo/packr/v2"

	"github.com/labstack/echo/v4"
)

// RunServer runs the main echo server
func RunServer(config config.Config) (err error) {
	var p page.Page

	templatesbox := packr.New("templates", "../../templates")
	staticbox := packr.New("static", "../../static")

	builtTemplates, _ := templates.BuildTemplates(templatesbox)

	e := echo.New()
	e.HideBanner = true
	e.Renderer = builtTemplates

	e.GET("/", p.Index)
	e.GET("/static/*", func(c echo.Context) error {
		return static.GetStatic(staticbox, c)
	})
	e.POST("/qrzws", func(c echo.Context) (err error) {
		res, err := Run(c, config)
		if err != nil {
			return c.String(http.StatusInternalServerError, "Erreur 500 ta mère")
		}
		return c.JSON(http.StatusOK, res)
	})
	/*e.GET("/export_frs.csv", func(c echo.Context) (err error) {
		data, mime, err := RunCSVExport(c, config)
		if err != nil {
			log.Fatalln(err)
		}
		return c.Blob(http.StatusOK, mime, data)
	})*/
	e.GET("/test", func(c echo.Context) (err error) {
		res, err := Test(c, config)
		if err != nil {
			return c.String(http.StatusInternalServerError, fmt.Sprintf("Erreur 500 ta mère: %d", res))
		}
		return c.JSON(http.StatusOK, res)
	})

	if !config.NoFeed {
		go qrz.Run(config)
	}
	e.Logger.Fatal(e.Start(fmt.Sprintf(":%d", config.Port)))
	return
}

// Test runs the main loop
func Test(c echo.Context, config config.Config) (cnt int64, err error) {
	//config.Db.Where("qrz LIKE '%%' OR name LIKE '%%' OR city LIKE '%%' OR dept LIKE '%%' OR country LIKE '%%' OR dmrid LIKE '%%'").Limit(50, 2).Desc("id").Select("*").Find(&qrz)
	cnt, err = config.Db.SQL("select count(*) from qrz").Count()

	return
}

// Run runs the main loop
func Run(c echo.Context, config config.Config) (res QrzDatatableOutput, err error) {
	qrzinputjson := new(QrzDatatableInput)

	if err = c.Bind(qrzinputjson); err != nil {
		return
	}

	rows, err := BuildQuery(config, *qrzinputjson)
	if err != nil {
		fmt.Println(err)
		return
	}

	for _, b := range rows {
		res.Data = append(res.Data, b)
	}

	res.RecordsFiltered, err = BuildQueryCountFiltered(config, *qrzinputjson)
	if err != nil {
		fmt.Println(err, "filtered")
		return
	}

	res.RecordsTotal, err = BuildQueryCountTotal(config, *qrzinputjson)
	if err != nil {
		fmt.Println(err, "total")
		return
	}

	res.Draw = qrzinputjson.Draw

	return
}

// BuildQuery builds main query
func BuildQuery(config config.Config, qrzdt QrzDatatableInput) (qrz []qrz.Qrz, err error) {
	selectstatement, err := SetSelect(config, qrzdt)
	if err != nil {
		return nil, err
	}

	orderdir, ordercol, err := SetOrder(config, qrzdt)
	if err != nil {
		return nil, err
	}

	limit, offset, err := SetLimit(config, qrzdt)
	if err != nil {
		return nil, err
	}

	searchstatement, err := SetSearchLike(config, qrzdt)
	if err != nil {
		return nil, err
	}

	if orderdir == "DESC" {
		err = config.Db.Where(searchstatement).Limit(limit, offset).Desc(ordercol...).Select(selectstatement).Find(&qrz)
	} else if orderdir == "ASC" {
		err = config.Db.Where(searchstatement).Limit(limit, offset).Asc(ordercol...).Select(selectstatement).Find(&qrz)
	} else {
		err = config.Db.Where(searchstatement).Limit(limit, offset).Asc("qrz").Select(selectstatement).Find(&qrz)
	}

	return
}

// BuildQueryCountFiltered builds query for counting filtered
func BuildQueryCountFiltered(config config.Config, qrzdt QrzDatatableInput) (cnt int64, err error) {
	searchstatement, err := SetSearchLike(config, qrzdt)
	cnt, err = config.Db.SQL(fmt.Sprintf("select count(*) from qrz WHERE %s", searchstatement)).Count()

	return
}

// BuildQueryCountTotal builds query for counting totals
func BuildQueryCountTotal(config config.Config, qrzdt QrzDatatableInput) (cnt int64, err error) {
	cnt, err = config.Db.SQL("select count(*) from qrz").Count()

	return
}

// SetSelect build the sql select statement part
func SetSelect(config config.Config, qrzdt QrzDatatableInput) (selectstatement string, err error) {
	var cols []string
	colre := regexp.MustCompile(`^[A-Za-z0-9]+$`)

	if len(qrzdt.Columns) > 0 {
		for _, col := range qrzdt.Columns {
			valid := colre.MatchString(col.Name)
			if valid {
				cols = append(cols, col.Name)
			} else {
				err = fmt.Errorf("String does not match prerequisites")
				return
			}
		}
		selectstatement = strings.Join(cols, ",")
	} else {
		selectstatement = "*"
	}

	return
}

// SetOrder build the sql order statement part
func SetOrder(config config.Config, qrzdt QrzDatatableInput) (orderdir string, ordercols []string, err error) {
	colre := regexp.MustCompile(`^[A-Za-z0-9]+$`)
	orderre := regexp.MustCompile(`^(ASC|asc|DESC|desc)$`)

	for _, col := range qrzdt.Order {
		if colre.MatchString(qrzdt.Columns[col.Column].Name) && orderre.MatchString(col.Dir) {
			ordercols = append(ordercols, fmt.Sprintf("%s", qrzdt.Columns[col.Column].Name))
			orderdir = col.Dir
		} else {
			err = fmt.Errorf("Order statements does not match prerequisites")
			return
		}
	}

	return
}

// SetLimit build the sql limit
func SetLimit(config config.Config, qrzdt QrzDatatableInput) (limit int, offset int, err error) {
	intre := regexp.MustCompile(`^[0-9]+$`)

	if qrzdt.Length < 1 {
		qrzdt.Length = 50
	}

	if intre.MatchString(fmt.Sprintf("%d", qrzdt.Length)) && intre.MatchString(fmt.Sprintf("%d", qrzdt.Start)) {
		limit = qrzdt.Length
		offset = qrzdt.Start
	} else {
		err = fmt.Errorf("Limit statements does not match prerequisites")
		return
	}

	return
}

// SetSearchLike build the where clause in sql statement
func SetSearchLike(config config.Config, qrzdt QrzDatatableInput) (searchstmt string, err error) {
	var searchstmtslice []string
	if len(qrzdt.Columns) > 0 {
		for id, i := range qrzdt.Columns {
			searchstmtslice = append(searchstmtslice, fmt.Sprintf("%s LIKE '%%%s%%'", i.Name, qrzdt.Search.Value))
			if id < len(qrzdt.Columns)-1 {
				searchstmtslice = append(searchstmtslice, "OR")
			}
		}

	} else {
		searchstmtslice = []string{"1=1"}
	}

	searchstmt = strings.Join(searchstmtslice, " ")

	return
}

/*
// RunCSVExport runs the main loop
func RunCSVExport(c echo.Context, config config.Config) (data []byte, mime string, err error) {
	mime = "text/csv"
	res, err := config.Db.Exec(config.DbStatements.ExportCSV)
	var res []string
	res = append(res, "QRZ,Name,City,Dept,Country")
	for rows.Next() {
		var l []string
		results := make(map[string]interface{})
		err = rows.MapScan(results)
		colslice, _ := rows.Columns()
		for _, column := range colslice {
			if results[column] == nil {
				l = append(l, "")
			} else {
				l = append(l, fmt.Sprintf("%s", results[column]))
			}
		}
		line := strings.Join(l, ",")
		res = append(res, line)
	}
	data = []byte(strings.Join(res, "\n"))

	return
}
*/

// QrzDatatableInput describes datatable query
type QrzDatatableInput struct {
	Columns []struct {
		Data       int    `json:"data"`
		Name       string `json:"name"`
		Searchable bool   `json:"searchable"`
		Orderable  bool   `json:"orderable"`
		Search     struct {
			Regex bool   `json:"regex"`
			Value string `json:"value"`
		} `json:"search"`
	} `json:"columns"`
	Draw   int `json:"draw"`
	Length int `json:"length"`
	Order  []struct {
		Column int    `json:"column"`
		Dir    string `json:"dir"`
	} `json:"order"`
	Search struct {
		Regex bool   `json:"regex"`
		Value string `json:"value"`
	} `json:"search"`
	Start int `json:"start"`
}

// QrzDatatableOutput describes datatable response
type QrzDatatableOutput struct {
	Data            []qrz.Qrz `json:"data"`
	Draw            int       `json:"draw"`
	RecordsTotal    int64     `json:"recordsTotal"`
	RecordsFiltered int64     `json:"recordsFiltered"`
}
